@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 1994-1996,1998-2011 Todd C. Miller <Todd.Miller@courtesan.com>
+ * Copyright (c) 1994-1996,1998-2015 Todd C. Miller <Todd.Miller@courtesan.com>
  *
  * Permission to use, copy, modify, and distribute this software for any
  * purpose with or without fee is hereby granted, provided that the above
@@ -21,26 +21,16 @@
 #include <config.h>
 
 #include <sys/types.h>
-#include <sys/param.h>
 #include <sys/stat.h>
 #include <stdio.h>
-#ifdef STDC_HEADERS
-# include <stdlib.h>
-# include <stddef.h>
-#else
-# ifdef HAVE_STDLIB_H
-#  include <stdlib.h>
-# endif
-#endif /* STDC_HEADERS */
+#include <stdlib.h>
 #ifdef HAVE_STRING_H
 # include <string.h>
 #endif /* HAVE_STRING_H */
 #ifdef HAVE_STRINGS_H
 # include <strings.h>
 #endif /* HAVE_STRINGS_H */
-#ifdef HAVE_UNISTD_H
-# include <unistd.h>
-#endif /* HAVE_UNISTD_H */
+#include <unistd.h>
 #ifdef _AIX
 # include <sys/id.h>
 #endif
@@ -80,20 +70,24 @@ static struct perm_state perm_stack[PERM_STACK_MAX];
 static int perm_stack_depth = 0;
 
 #undef ID
-#define ID(x) (state->x == ostate->x ? -1 : state->x)
+#define ID(x) (state->x == ostate->x ? (uid_t)-1 : state->x)
 #undef OID
-#define OID(x) (ostate->x == state->x ? -1 : ostate->x)
+#define OID(x) (ostate->x == state->x ? (uid_t)-1 : ostate->x)
 
-void
+bool
 rewind_perms(void)
 {
-    debug_decl(rewind_perms, SUDO_DEBUG_PERMS)
+    debug_decl(rewind_perms, SUDOERS_DEBUG_PERMS)
 
-    while (perm_stack_depth > 1)
-	restore_perms();
-    grlist_delref(perm_stack[0].grlist);
+    if (perm_stack_depth != 0) {
+	while (perm_stack_depth > 1) {
+	    if (!restore_perms())
+		debug_return_bool(false);
+	}
+	sudo_grlist_delref(perm_stack[0].grlist);
+    }
 
-    debug_return;
+    debug_return_bool(true);
 }
 
 #if defined(HAVE_SETRESUID)
@@ -107,19 +101,16 @@ rewind_perms(void)
  * We only flip the effective gid since it only changes for PERM_SUDOERS.
  * This version of set_perms() works fine with the "stay_setuid" option.
  */
-int
+bool
 set_perms(int perm)
 {
     struct perm_state *state, *ostate = NULL;
     char errbuf[1024];
-    int noexit;
-    debug_decl(set_perms, SUDO_DEBUG_PERMS)
-
-    noexit = ISSET(perm, PERM_NOEXIT);
-    CLR(perm, PERM_MASK);
+    const char *errstr = errbuf;
+    debug_decl(set_perms, SUDOERS_DEBUG_PERMS)
 
     if (perm_stack_depth == PERM_STACK_MAX) {
-	strlcpy(errbuf, _("perm stack overflow"), sizeof(errbuf));
+	errstr = N_("perm stack overflow");
 	errno = EINVAL;
 	goto bad;
     }
@@ -127,7 +118,7 @@ set_perms(int perm)
     state = &perm_stack[perm_stack_depth];
     if (perm != PERM_INITIAL) {
 	if (perm_stack_depth == 0) {
-	    strlcpy(errbuf, _("perm stack underflow"), sizeof(errbuf));
+	    errstr = N_("perm stack underflow");
 	    errno = EINVAL;
 	    goto bad;
 	}
@@ -139,12 +130,12 @@ set_perms(int perm)
 	/* Stash initial state */
 #ifdef HAVE_GETRESUID
 	if (getresuid(&state->ruid, &state->euid, &state->suid)) {
-	    strlcpy(errbuf, "PERM_INITIAL: getresuid", sizeof(errbuf));
+	    errstr = "PERM_INITIAL: getresuid";
 	    goto bad;
 
 	}
 	if (getresgid(&state->rgid, &state->egid, &state->sgid)) {
-	    strlcpy(errbuf, "PERM_INITIAL: getresgid", sizeof(errbuf));
+	    errstr = "PERM_INITIAL: getresgid";
 	    goto bad;
 	}
 #else
@@ -157,7 +148,7 @@ set_perms(int perm)
 	state->sgid = state->egid; /* in case we are setgid */
 #endif
 	state->grlist = user_group_list;
-	grlist_addref(state->grlist);
+	sudo_grlist_addref(state->grlist);
 	sudo_debug_printf(SUDO_DEBUG_INFO, "%s: PERM_INITIAL: "
 	    "ruid: %d, euid: %d, suid: %d, rgid: %d, egid: %d, sgid: %d",
 	    __func__, (int)state->ruid, (int)state->euid, (int)state->suid,
@@ -179,10 +170,18 @@ set_perms(int perm)
 	    goto bad;
 	}
 	state->rgid = ostate->rgid;
-	state->egid = ostate->egid;
+	state->egid = ROOT_GID;
 	state->sgid = ostate->sgid;
+	sudo_debug_printf(SUDO_DEBUG_INFO, "%s: PERM_ROOT: gid: "
+	    "[%d, %d, %d] -> [%d, %d, %d]", __func__,
+	    (int)ostate->rgid, (int)ostate->egid, (int)ostate->sgid,
+	    (int)state->rgid, (int)state->egid, (int)state->sgid);
+	if (GID_CHANGED && setresgid(ID(rgid), ID(egid), ID(sgid))) {
+	    errstr = N_("unable to change to root gid");
+	    goto bad;
+	}
 	state->grlist = ostate->grlist;
-	grlist_addref(state->grlist);
+	sudo_grlist_addref(state->grlist);
 	break;
 
     case PERM_USER:
@@ -199,10 +198,10 @@ set_perms(int perm)
 	    goto bad;
 	}
 	state->grlist = user_group_list;
-	grlist_addref(state->grlist);
+	sudo_grlist_addref(state->grlist);
 	if (state->grlist != ostate->grlist) {
 	    if (sudo_setgroups(state->grlist->ngids, state->grlist->gids)) {
-		strlcpy(errbuf, "PERM_USER: setgroups", sizeof(errbuf));
+		errstr = "PERM_USER: setgroups";
 		goto bad;
 	    }
 	}
@@ -236,10 +235,10 @@ set_perms(int perm)
 	    goto bad;
 	}
 	state->grlist = user_group_list;
-	grlist_addref(state->grlist);
+	sudo_grlist_addref(state->grlist);
 	if (state->grlist != ostate->grlist) {
 	    if (sudo_setgroups(state->grlist->ngids, state->grlist->gids)) {
-		strlcpy(errbuf, "PERM_FULL_USER: setgroups", sizeof(errbuf));
+		errstr = "PERM_FULL_USER: setgroups";
 		goto bad;
 	    }
 	}
@@ -267,10 +266,14 @@ set_perms(int perm)
 	    (int)ostate->rgid, (int)ostate->egid, (int)ostate->sgid,
 	    (int)state->rgid, (int)state->egid, (int)state->sgid);
 	if (GID_CHANGED && setresgid(ID(rgid), ID(egid), ID(sgid))) {
-	    strlcpy(errbuf, _("unable to change to runas gid"), sizeof(errbuf));
+	    errstr = N_("unable to change to runas gid");
 	    goto bad;
 	}
 	state->grlist = runas_setgroups();
+	if (state->grlist == NULL) {
+	    errstr = N_("unable to set runas group vector");
+	    goto bad;
+	}
 	state->ruid = ostate->ruid;
 	state->euid = runas_pw ? runas_pw->pw_uid : user_uid;
 	state->suid = ostate->suid;
@@ -279,14 +282,14 @@ set_perms(int perm)
 	    (int)ostate->ruid, (int)ostate->euid, (int)ostate->suid,
 	    (int)state->ruid, (int)state->euid, (int)state->suid);
 	if (UID_CHANGED && setresuid(ID(ruid), ID(euid), ID(suid))) {
-	    strlcpy(errbuf, _("unable to change to runas uid"), sizeof(errbuf));
+	    errstr = N_("unable to change to runas uid");
 	    goto bad;
 	}
 	break;
 
     case PERM_SUDOERS:
 	state->grlist = ostate->grlist;
-	grlist_addref(state->grlist);
+	sudo_grlist_addref(state->grlist);
 
 	/* assumes euid == ROOT_UID, ruid == user */
 	state->rgid = ostate->rgid;
@@ -297,7 +300,7 @@ set_perms(int perm)
 	    (int)ostate->rgid, (int)ostate->egid, (int)ostate->sgid,
 	    (int)state->rgid, (int)state->egid, (int)state->sgid);
 	if (GID_CHANGED && setresgid(ID(rgid), ID(egid), ID(sgid))) {
-	    strlcpy(errbuf, _("unable to change to sudoers gid"), sizeof(errbuf));
+	    errstr = N_("unable to change to sudoers gid");
 	    goto bad;
 	}
 
@@ -326,7 +329,7 @@ set_perms(int perm)
 
     case PERM_TIMESTAMP:
 	state->grlist = ostate->grlist;
-	grlist_addref(state->grlist);
+	sudo_grlist_addref(state->grlist);
 	state->rgid = ostate->rgid;
 	state->egid = ostate->egid;
 	state->sgid = ostate->sgid;
@@ -347,23 +350,25 @@ set_perms(int perm)
     }
 
     perm_stack_depth++;
-    debug_return_bool(1);
+    debug_return_bool(true);
 bad:
-    warningx("%s: %s", errbuf,
-	errno == EAGAIN ? _("too many processes") : strerror(errno));
-    if (noexit)
-	debug_return_bool(0);
-    exit(1);
+    if (errno == EAGAIN)
+	sudo_warnx(U_("%s: %s"), U_(errstr), U_("too many processes"));
+    else
+	sudo_warn("%s", U_(errstr));
+    debug_return_bool(false);
 }
 
-void
+bool
 restore_perms(void)
 {
     struct perm_state *state, *ostate;
-    debug_decl(restore_perms, SUDO_DEBUG_PERMS)
+    debug_decl(restore_perms, SUDOERS_DEBUG_PERMS)
 
-    if (perm_stack_depth < 2)
-	debug_return;
+    if (perm_stack_depth < 2) {
+	sudo_warnx(U_("perm stack underflow"));
+	debug_return_bool(true);
+    }
 
     state = &perm_stack[perm_stack_depth - 1];
     ostate = &perm_stack[perm_stack_depth - 2];
@@ -379,35 +384,35 @@ restore_perms(void)
     /* XXX - more cases here where euid != ruid */
     if (OID(euid) == ROOT_UID) {
 	if (setresuid(-1, ROOT_UID, -1)) {
-	    warning("setresuid() [%d, %d, %d] -> [%d, %d, %d]",
+	    sudo_warn("setresuid() [%d, %d, %d] -> [%d, %d, %d]",
 		(int)state->ruid, (int)state->euid, (int)state->suid,
 		-1, ROOT_UID, -1);
 	    goto bad;
 	}
     }
     if (setresuid(OID(ruid), OID(euid), OID(suid))) {
-	warning("setresuid() [%d, %d, %d] -> [%d, %d, %d]",
+	sudo_warn("setresuid() [%d, %d, %d] -> [%d, %d, %d]",
 	    (int)state->ruid, (int)state->euid, (int)state->suid,
 	    (int)OID(ruid), (int)OID(euid), (int)OID(suid));
 	goto bad;
     }
     if (setresgid(OID(rgid), OID(egid), OID(sgid))) {
-	warning("setresgid() [%d, %d, %d] -> [%d, %d, %d]",
+	sudo_warn("setresgid() [%d, %d, %d] -> [%d, %d, %d]",
 	    (int)state->rgid, (int)state->egid, (int)state->sgid,
 	    (int)OID(rgid), (int)OID(egid), (int)OID(sgid));
 	goto bad;
     }
     if (state->grlist != ostate->grlist) {
 	if (sudo_setgroups(ostate->grlist->ngids, ostate->grlist->gids)) {
-	    warning("setgroups()");
+	    sudo_warn("setgroups()");
 	    goto bad;
 	}
     }
-    grlist_delref(state->grlist);
-    debug_return;
+    sudo_grlist_delref(state->grlist);
+    debug_return_bool(true);
 
 bad:
-    exit(1);
+    debug_return_bool(false);
 }
 
 #elif defined(_AIX) && defined(ID_SAVED)
@@ -421,19 +426,16 @@ bad:
  * We only flip the effective gid since it only changes for PERM_SUDOERS.
  * This version of set_perms() works fine with the "stay_setuid" option.
  */
-int
+bool
 set_perms(int perm)
 {
     struct perm_state *state, *ostate = NULL;
     char errbuf[1024];
-    int noexit;
-    debug_decl(set_perms, SUDO_DEBUG_PERMS)
-
-    noexit = ISSET(perm, PERM_NOEXIT);
-    CLR(perm, PERM_MASK);
+    const char *errstr = errbuf;
+    debug_decl(set_perms, SUDOERS_DEBUG_PERMS)
 
     if (perm_stack_depth == PERM_STACK_MAX) {
-	strlcpy(errbuf, _("perm stack overflow"), sizeof(errbuf));
+	errstr = N_("perm stack overflow");
 	errno = EINVAL;
 	goto bad;
     }
@@ -441,7 +443,7 @@ set_perms(int perm)
     state = &perm_stack[perm_stack_depth];
     if (perm != PERM_INITIAL) {
 	if (perm_stack_depth == 0) {
-	    strlcpy(errbuf, _("perm stack underflow"), sizeof(errbuf));
+	    errstr = N_("perm stack underflow");
 	    errno = EINVAL;
 	    goto bad;
 	}
@@ -458,7 +460,7 @@ set_perms(int perm)
 	state->egid = getgidx(ID_EFFECTIVE);
 	state->sgid = getgidx(ID_SAVED);
 	state->grlist = user_group_list;
-	grlist_addref(state->grlist);
+	sudo_grlist_addref(state->grlist);
 	sudo_debug_printf(SUDO_DEBUG_INFO, "%s: PERM_INITIAL: "
 	    "ruid: %d, euid: %d, suid: %d, rgid: %d, egid: %d, sgid: %d",
 	    __func__, (unsigned int)state->ruid, (unsigned int)state->euid,
@@ -481,10 +483,18 @@ set_perms(int perm)
 	    goto bad;
 	}
 	state->rgid = ostate->rgid;
-	state->egid = ostate->egid;
+	state->egid = ROOT_GID;
 	state->sgid = ostate->sgid;
+	sudo_debug_printf(SUDO_DEBUG_INFO, "%s: PERM_ROOT: gid: "
+	    "[%d, %d, %d] -> [%d, %d, %d]", __func__,
+	    (int)ostate->rgid, (int)ostate->egid, (int)ostate->sgid,
+	    (int)state->rgid, (int)state->egid, (int)state->sgid);
+	if (GID_CHANGED && setgidx(ID_EFFECTIVE, ROOT_GID)) {
+	    errstr = N_("unable to change to root gid");
+	    goto bad;
+	}
 	state->grlist = ostate->grlist;
-	grlist_addref(state->grlist);
+	sudo_grlist_addref(state->grlist);
 	break;
 
     case PERM_USER:
@@ -501,10 +511,10 @@ set_perms(int perm)
 	    goto bad;
 	}
 	state->grlist = user_group_list;
-	grlist_addref(state->grlist);
+	sudo_grlist_addref(state->grlist);
 	if (state->grlist != ostate->grlist) {
 	    if (sudo_setgroups(state->grlist->ngids, state->grlist->gids)) {
-		strlcpy(errbuf, "PERM_USER: setgroups", sizeof(errbuf));
+		errstr = "PERM_USER: setgroups";
 		goto bad;
 	    }
 	}
@@ -546,10 +556,10 @@ set_perms(int perm)
 	    goto bad;
 	}
 	state->grlist = user_group_list;
-	grlist_addref(state->grlist);
+	sudo_grlist_addref(state->grlist);
 	if (state->grlist != ostate->grlist) {
 	    if (sudo_setgroups(state->grlist->ngids, state->grlist->gids)) {
-		strlcpy(errbuf, "PERM_FULL_USER: setgroups", sizeof(errbuf));
+		errstr = "PERM_FULL_USER: setgroups";
 		goto bad;
 	    }
 	}
@@ -577,10 +587,14 @@ set_perms(int perm)
 	    (int)ostate->rgid, (int)ostate->egid, (int)ostate->sgid,
 	    (int)state->rgid, (int)state->egid, (int)state->sgid);
 	if (GID_CHANGED && setgidx(ID_EFFECTIVE, state->egid)) {
-	    strlcpy(errbuf, _("unable to change to runas gid"), sizeof(errbuf));
+	    errstr = N_("unable to change to runas gid");
 	    goto bad;
 	}
 	state->grlist = runas_setgroups();
+	if (state->grlist == NULL) {
+	    errstr = N_("unable to set runas group vector");
+	    goto bad;
+	}
 	state->ruid = ostate->ruid;
 	state->euid = runas_pw ? runas_pw->pw_uid : user_uid;
 	state->suid = ostate->suid;
@@ -589,14 +603,14 @@ set_perms(int perm)
 	    (int)ostate->ruid, (int)ostate->euid, (int)ostate->suid,
 	    (int)state->ruid, (int)state->euid, (int)state->suid);
 	if (UID_CHANGED && setuidx(ID_EFFECTIVE, state->euid)) {
-	    strlcpy(errbuf, _("unable to change to runas uid"), sizeof(errbuf));
+	    errstr = N_("unable to change to runas uid");
 	    goto bad;
 	}
 	break;
 
     case PERM_SUDOERS:
 	state->grlist = ostate->grlist;
-	grlist_addref(state->grlist);
+	sudo_grlist_addref(state->grlist);
 
 	/* assume euid == ROOT_UID, ruid == user */
 	state->rgid = ostate->rgid;
@@ -607,7 +621,7 @@ set_perms(int perm)
 	    (int)ostate->rgid, (int)ostate->egid, (int)ostate->sgid,
 	    (int)state->rgid, (int)state->egid, (int)state->sgid);
 	if (GID_CHANGED && setgidx(ID_EFFECTIVE, sudoers_gid)) {
-	    strlcpy(errbuf, _("unable to change to sudoers gid"), sizeof(errbuf));
+	    errstr = N_("unable to change to sudoers gid");
 	    goto bad;
 	}
 
@@ -645,7 +659,7 @@ set_perms(int perm)
 
     case PERM_TIMESTAMP:
 	state->grlist = ostate->grlist;
-	grlist_addref(state->grlist);
+	sudo_grlist_addref(state->grlist);
 	state->rgid = ostate->rgid;
 	state->egid = ostate->egid;
 	state->sgid = ostate->sgid;
@@ -675,23 +689,25 @@ set_perms(int perm)
     }
 
     perm_stack_depth++;
-    debug_return_bool(1);
+    debug_return_bool(true);
 bad:
-    warningx("%s: %s", errbuf,
-	errno == EAGAIN ? _("too many processes") : strerror(errno));
-    if (noexit)
-	debug_return_bool(0);
-    exit(1);
+    if (errno == EAGAIN)
+	sudo_warnx(U_("%s: %s"), U_(errstr), U_("too many processes"));
+    else
+	sudo_warn("%s", U_(errstr));
+    debug_return_bool(false);
 }
 
-void
+bool
 restore_perms(void)
 {
     struct perm_state *state, *ostate;
-    debug_decl(restore_perms, SUDO_DEBUG_PERMS)
+    debug_decl(restore_perms, SUDOERS_DEBUG_PERMS)
 
-    if (perm_stack_depth < 2)
-	debug_return;
+    if (perm_stack_depth < 2) {
+	sudo_warnx(U_("perm stack underflow"));
+	debug_return_bool(true);
+    }
 
     state = &perm_stack[perm_stack_depth - 1];
     ostate = &perm_stack[perm_stack_depth - 2];
@@ -704,12 +720,12 @@ restore_perms(void)
 	__func__, (int)state->rgid, (int)state->egid, (int)state->sgid,
 	(int)ostate->rgid, (int)ostate->egid, (int)ostate->sgid);
 
-    if (OID(ruid) != -1 || OID(euid) != -1 || OID(suid) != -1) {
+    if (OID(ruid) != (uid_t)-1 || OID(euid) != (uid_t)-1 || OID(suid) != (uid_t)-1) {
 	if (OID(euid) == ROOT_UID) {
 	    sudo_debug_printf(SUDO_DEBUG_INFO, "%s: setuidx(ID_EFFECTIVE, %d)",
 		__func__, ROOT_UID);
 	    if (setuidx(ID_EFFECTIVE, ROOT_UID)) {
-		warning("setuidx(ID_EFFECTIVE) [%d, %d, %d] -> [%d, %d, %d]",
+		sudo_warn("setuidx(ID_EFFECTIVE) [%d, %d, %d] -> [%d, %d, %d]",
 		    (int)state->ruid, (int)state->euid, (int)state->suid,
 		    -1, ROOT_UID, -1);
 		goto bad;
@@ -720,29 +736,29 @@ restore_perms(void)
 		"%s: setuidx(ID_EFFECTIVE|ID_REAL|ID_SAVED, %d)",
 		__func__, OID(ruid));
 	    if (setuidx(ID_EFFECTIVE|ID_REAL|ID_SAVED, OID(ruid))) {
-		warning("setuidx(ID_EFFECTIVE|ID_REAL|ID_SAVED) [%d, %d, %d] -> [%d, %d, %d]",
+		sudo_warn("setuidx(ID_EFFECTIVE|ID_REAL|ID_SAVED) [%d, %d, %d] -> [%d, %d, %d]",
 		    (int)state->ruid, (int)state->euid, (int)state->suid,
 		    (int)OID(ruid), (int)OID(euid), (int)OID(suid));
 		goto bad;
 	    }
-	} else if (OID(ruid) == -1 && OID(suid) == -1) {
+	} else if (OID(ruid) == (uid_t)-1 && OID(suid) == (uid_t)-1) {
 	    /* May have already changed euid to ROOT_UID above. */
 	    if (OID(euid) != ROOT_UID) {
 		sudo_debug_printf(SUDO_DEBUG_INFO,
 		    "%s: setuidx(ID_EFFECTIVE, %d)", __func__, OID(euid));
 		if (setuidx(ID_EFFECTIVE, OID(euid))) {
-		    warning("setuidx(ID_EFFECTIVE) [%d, %d, %d] -> [%d, %d, %d]",
+		    sudo_warn("setuidx(ID_EFFECTIVE) [%d, %d, %d] -> [%d, %d, %d]",
 			(int)state->ruid, (int)state->euid, (int)state->suid,
 			(int)OID(ruid), (int)OID(euid), (int)OID(suid));
 		    goto bad;
 		}
 	    }
-	} else if (OID(suid) == -1) {
+	} else if (OID(suid) == (uid_t)-1) {
 	    /* Cannot set the real uid alone. */
 	    sudo_debug_printf(SUDO_DEBUG_INFO,
 		"%s: setuidx(ID_REAL|ID_EFFECTIVE, %d)", __func__, OID(ruid));
 	    if (setuidx(ID_REAL|ID_EFFECTIVE, OID(ruid))) {
-		warning("setuidx(ID_REAL|ID_EFFECTIVE) [%d, %d, %d] -> [%d, %d, %d]",
+		sudo_warn("setuidx(ID_REAL|ID_EFFECTIVE) [%d, %d, %d] -> [%d, %d, %d]",
 		    (int)state->ruid, (int)state->euid, (int)state->suid,
 		    (int)OID(ruid), (int)OID(euid), (int)OID(suid));
 		goto bad;
@@ -752,37 +768,37 @@ restore_perms(void)
 		sudo_debug_printf(SUDO_DEBUG_INFO,
 		    "%s: setuidx(ID_EFFECTIVE, %d)", __func__, ostate->euid);
 		if (setuidx(ID_EFFECTIVE, ostate->euid)) {
-		    warning("setuidx(ID_EFFECTIVE, %d)", ostate->euid);
+		    sudo_warn("setuidx(ID_EFFECTIVE, %d)", ostate->euid);
 		    goto bad;
 		}
 	    }
 	}
     }
-    if (OID(rgid) != -1 || OID(egid) != -1 || OID(sgid) != -1) {
+    if (OID(rgid) != (gid_t)-1 || OID(egid) != (gid_t)-1 || OID(sgid) != (gid_t)-1) {
 	if (OID(rgid) == OID(egid) && OID(egid) == OID(sgid)) {
 	    sudo_debug_printf(SUDO_DEBUG_INFO,
 		"%s: setgidx(ID_EFFECTIVE|ID_REAL|ID_SAVED, %d)",
 		__func__, OID(rgid));
 	    if (setgidx(ID_EFFECTIVE|ID_REAL|ID_SAVED, OID(rgid))) {
-		warning("setgidx(ID_EFFECTIVE|ID_REAL|ID_SAVED) [%d, %d, %d] -> [%d, %d, %d]",
+		sudo_warn("setgidx(ID_EFFECTIVE|ID_REAL|ID_SAVED) [%d, %d, %d] -> [%d, %d, %d]",
 		    (int)state->rgid, (int)state->egid, (int)state->sgid,
 		    (int)OID(rgid), (int)OID(egid), (int)OID(sgid));
 		goto bad;
 	    }
-	} else if (OID(rgid) == -1 && OID(sgid) == -1) {
+	} else if (OID(rgid) == (gid_t)-1 && OID(sgid) == (gid_t)-1) {
 	    sudo_debug_printf(SUDO_DEBUG_INFO, "%s: setgidx(ID_EFFECTIVE, %d)",
 		__func__, OID(egid));
 	    if (setgidx(ID_EFFECTIVE, OID(egid))) {
-		warning("setgidx(ID_EFFECTIVE) [%d, %d, %d] -> [%d, %d, %d]",
+		sudo_warn("setgidx(ID_EFFECTIVE) [%d, %d, %d] -> [%d, %d, %d]",
 		    (int)state->rgid, (int)state->egid, (int)state->sgid,
 		    (int)OID(rgid), (int)OID(egid), (int)OID(sgid));
 		goto bad;
 	    }
-	} else if (OID(sgid) == -1) {
+	} else if (OID(sgid) == (gid_t)-1) {
 	    sudo_debug_printf(SUDO_DEBUG_INFO,
 		"%s: setgidx(ID_EFFECTIVE|ID_REAL, %d)", __func__, OID(rgid));
 	    if (setgidx(ID_REAL|ID_EFFECTIVE, OID(rgid))) {
-		warning("setgidx(ID_REAL|ID_EFFECTIVE) [%d, %d, %d] -> [%d, %d, %d]",
+		sudo_warn("setgidx(ID_REAL|ID_EFFECTIVE) [%d, %d, %d] -> [%d, %d, %d]",
 		    (int)state->rgid, (int)state->egid, (int)state->sgid,
 		    (int)OID(rgid), (int)OID(egid), (int)OID(sgid));
 		goto bad;
@@ -791,15 +807,15 @@ restore_perms(void)
     }
     if (state->grlist != ostate->grlist) {
 	if (sudo_setgroups(ostate->grlist->ngids, ostate->grlist->gids)) {
-	    warning("setgroups()");
+	    sudo_warn("setgroups()");
 	    goto bad;
 	}
     }
-    grlist_delref(state->grlist);
-    debug_return;
+    sudo_grlist_delref(state->grlist);
+    debug_return_bool(true);
 
 bad:
-    exit(1);
+    debug_return_bool(false);
 }
 
 #elif defined(HAVE_SETREUID)
@@ -813,19 +829,16 @@ bad:
  * We only flip the effective gid since it only changes for PERM_SUDOERS.
  * This version of set_perms() works fine with the "stay_setuid" option.
  */
-int
+bool
 set_perms(int perm)
 {
     struct perm_state *state, *ostate = NULL;
     char errbuf[1024];
-    int noexit;
-    debug_decl(set_perms, SUDO_DEBUG_PERMS)
-
-    noexit = ISSET(perm, PERM_NOEXIT);
-    CLR(perm, PERM_MASK);
+    const char *errstr = errbuf;
+    debug_decl(set_perms, SUDOERS_DEBUG_PERMS)
 
     if (perm_stack_depth == PERM_STACK_MAX) {
-	strlcpy(errbuf, _("perm stack overflow"), sizeof(errbuf));
+	errstr = N_("perm stack overflow");
 	errno = EINVAL;
 	goto bad;
     }
@@ -833,7 +846,7 @@ set_perms(int perm)
     state = &perm_stack[perm_stack_depth];
     if (perm != PERM_INITIAL) {
 	if (perm_stack_depth == 0) {
-	    strlcpy(errbuf, _("perm stack underflow"), sizeof(errbuf));
+	    errstr = N_("perm stack underflow");
 	    errno = EINVAL;
 	    goto bad;
 	}
@@ -848,7 +861,7 @@ set_perms(int perm)
 	state->rgid = getgid();
 	state->egid = getegid();
 	state->grlist = user_group_list;
-	grlist_addref(state->grlist);
+	sudo_grlist_addref(state->grlist);
 	sudo_debug_printf(SUDO_DEBUG_INFO, "%s: PERM_INITIAL: "
 	    "ruid: %d, euid: %d, rgid: %d, egid: %d", __func__,
 	    (int)state->ruid, (int)state->euid,
@@ -867,7 +880,7 @@ set_perms(int perm)
 	if (ostate->euid != ROOT_UID) {
 	    if (setreuid(-1, ROOT_UID)) {
 		snprintf(errbuf, sizeof(errbuf),
-		    "PERM_ROOT: setreuid(-1, %d)", PERM_ROOT);
+		    "PERM_ROOT: setreuid(-1, %d)", ROOT_UID);
 		goto bad;
 	    }
 	}
@@ -879,9 +892,17 @@ set_perms(int perm)
 	    }
 	}
 	state->rgid = ostate->rgid;
-	state->egid = ostate->rgid;
+	state->egid = ROOT_GID;
+	sudo_debug_printf(SUDO_DEBUG_INFO, "%s: PERM_ROOT: gid: "
+	    "[%d, %d] -> [%d, %d]", __func__, (int)ostate->rgid,
+	    (int)ostate->egid, (int)state->rgid, (int)state->egid);
+	if (GID_CHANGED && setregid(ID(rgid), ID(egid))) {
+	    snprintf(errbuf, sizeof(errbuf),
+		"PERM_ROOT: setregid(%d, %d)", ID(rgid), ID(egid));
+	    goto bad;
+	}
 	state->grlist = ostate->grlist;
-	grlist_addref(state->grlist);
+	sudo_grlist_addref(state->grlist);
 	break;
 
     case PERM_USER:
@@ -896,10 +917,10 @@ set_perms(int perm)
 	    goto bad;
 	}
 	state->grlist = user_group_list;
-	grlist_addref(state->grlist);
+	sudo_grlist_addref(state->grlist);
 	if (state->grlist != ostate->grlist) {
 	    if (sudo_setgroups(state->grlist->ngids, state->grlist->gids)) {
-		strlcpy(errbuf, "PERM_USER: setgroups", sizeof(errbuf));
+		errstr = "PERM_USER: setgroups";
 		goto bad;
 	    }
 	}
@@ -928,10 +949,10 @@ set_perms(int perm)
 	    goto bad;
 	}
 	state->grlist = user_group_list;
-	grlist_addref(state->grlist);
+	sudo_grlist_addref(state->grlist);
 	if (state->grlist != ostate->grlist) {
 	    if (sudo_setgroups(state->grlist->ngids, state->grlist->gids)) {
-		strlcpy(errbuf, "PERM_FULL_USER: setgroups", sizeof(errbuf));
+		errstr = "PERM_FULL_USER: setgroups";
 		goto bad;
 	    }
 	}
@@ -954,24 +975,28 @@ set_perms(int perm)
 	    "[%d, %d] -> [%d, %d]", __func__, (int)ostate->rgid,
 	    (int)ostate->egid, (int)state->rgid, (int)state->egid);
 	if (GID_CHANGED && setregid(ID(rgid), ID(egid))) {
-	    strlcpy(errbuf, _("unable to change to runas gid"), sizeof(errbuf));
+	    errstr = N_("unable to change to runas gid");
 	    goto bad;
 	}
 	state->grlist = runas_setgroups();
+	if (state->grlist == NULL) {
+	    errstr = N_("unable to set runas group vector");
+	    goto bad;
+	}
 	state->ruid = ROOT_UID;
 	state->euid = runas_pw ? runas_pw->pw_uid : user_uid;
 	sudo_debug_printf(SUDO_DEBUG_INFO, "%s: PERM_RUNAS: uid: "
 	    "[%d, %d] -> [%d, %d]", __func__, (int)ostate->ruid,
 	    (int)ostate->euid, (int)state->ruid, (int)state->euid);
 	if (UID_CHANGED && setreuid(ID(ruid), ID(euid))) {
-	    strlcpy(errbuf, _("unable to change to runas uid"), sizeof(errbuf));
+	    errstr = N_("unable to change to runas uid");
 	    goto bad;
 	}
 	break;
 
     case PERM_SUDOERS:
 	state->grlist = ostate->grlist;
-	grlist_addref(state->grlist);
+	sudo_grlist_addref(state->grlist);
 
 	/* assume euid == ROOT_UID, ruid == user */
 	state->rgid = ostate->rgid;
@@ -980,7 +1005,7 @@ set_perms(int perm)
 	    "[%d, %d] -> [%d, %d]", __func__, (int)ostate->rgid,
 	    (int)ostate->egid, (int)state->rgid, (int)state->egid);
 	if (GID_CHANGED && setregid(ID(rgid), ID(egid))) {
-	    strlcpy(errbuf, _("unable to change to sudoers gid"), sizeof(errbuf));
+	    errstr = N_("unable to change to sudoers gid");
 	    goto bad;
 	}
 
@@ -1006,7 +1031,7 @@ set_perms(int perm)
 
     case PERM_TIMESTAMP:
 	state->grlist = ostate->grlist;
-	grlist_addref(state->grlist);
+	sudo_grlist_addref(state->grlist);
 	state->rgid = ostate->rgid;
 	state->egid = ostate->egid;
 	state->ruid = ROOT_UID;
@@ -1023,23 +1048,25 @@ set_perms(int perm)
     }
 
     perm_stack_depth++;
-    debug_return_bool(1);
+    debug_return_bool(true);
 bad:
-    warningx("%s: %s", errbuf,
-	errno == EAGAIN ? _("too many processes") : strerror(errno));
-    if (noexit)
-	debug_return_bool(0);
-    exit(1);
+    if (errno == EAGAIN)
+	sudo_warnx(U_("%s: %s"), U_(errstr), U_("too many processes"));
+    else
+	sudo_warn("%s", U_(errstr));
+    debug_return_bool(false);
 }
 
-void
+bool
 restore_perms(void)
 {
     struct perm_state *state, *ostate;
-    debug_decl(restore_perms, SUDO_DEBUG_PERMS)
+    debug_decl(restore_perms, SUDOERS_DEBUG_PERMS)
 
-    if (perm_stack_depth < 2)
-	debug_return;
+    if (perm_stack_depth < 2) {
+	sudo_warnx(U_("perm stack underflow"));
+	debug_return_bool(true);
+    }
 
     state = &perm_stack[perm_stack_depth - 1];
     ostate = &perm_stack[perm_stack_depth - 2];
@@ -1059,34 +1086,34 @@ restore_perms(void)
     if (OID(euid) == ROOT_UID) {
 	/* setuid() may not set the saved ID unless the euid is ROOT_UID */
 	if (ID(euid) != ROOT_UID)
-	    (void)setreuid(-1, ROOT_UID);
+	    ignore_result(setreuid(-1, ROOT_UID));
 	if (setuid(ROOT_UID)) {
-	    warning("setuid() [%d, %d] -> %d)", (int)state->ruid,
+	    sudo_warn("setuid() [%d, %d] -> %d)", (int)state->ruid,
 		(int)state->euid, ROOT_UID);
 	    goto bad;
 	}
     }
     if (setreuid(OID(ruid), OID(euid))) {
-	warning("setreuid() [%d, %d] -> [%d, %d]", (int)state->ruid,
+	sudo_warn("setreuid() [%d, %d] -> [%d, %d]", (int)state->ruid,
 	    (int)state->euid, (int)OID(ruid), (int)OID(euid));
 	goto bad;
     }
     if (setregid(OID(rgid), OID(egid))) {
-	warning("setregid() [%d, %d] -> [%d, %d]", (int)state->rgid,
+	sudo_warn("setregid() [%d, %d] -> [%d, %d]", (int)state->rgid,
 	    (int)state->egid, (int)OID(rgid), (int)OID(egid));
 	goto bad;
     }
     if (state->grlist != ostate->grlist) {
 	if (sudo_setgroups(ostate->grlist->ngids, ostate->grlist->gids)) {
-	    warning("setgroups()");
+	    sudo_warn("setgroups()");
 	    goto bad;
 	}
     }
-    grlist_delref(state->grlist);
-    debug_return;
+    sudo_grlist_delref(state->grlist);
+    debug_return_bool(true);
 
 bad:
-    exit(1);
+    debug_return_bool(false);
 }
 
 #elif defined(HAVE_SETEUID)
@@ -1099,19 +1126,16 @@ bad:
  * we are headed for an exec().
  * This version of set_perms() works fine with the "stay_setuid" option.
  */
-int
+bool
 set_perms(int perm)
 {
     struct perm_state *state, *ostate = NULL;
     char errbuf[1024];
-    int noexit;
-    debug_decl(set_perms, SUDO_DEBUG_PERMS)
-
-    noexit = ISSET(perm, PERM_NOEXIT);
-    CLR(perm, PERM_MASK);
+    const char *errstr = errbuf;
+    debug_decl(set_perms, SUDOERS_DEBUG_PERMS)
 
     if (perm_stack_depth == PERM_STACK_MAX) {
-	strlcpy(errbuf, _("perm stack overflow"), sizeof(errbuf));
+	errstr = N_("perm stack overflow");
 	errno = EINVAL;
 	goto bad;
     }
@@ -1119,7 +1143,7 @@ set_perms(int perm)
     state = &perm_stack[perm_stack_depth];
     if (perm != PERM_INITIAL) {
 	if (perm_stack_depth == 0) {
-	    strlcpy(errbuf, _("perm stack underflow"), sizeof(errbuf));
+	    errstr = N_("perm stack underflow");
 	    errno = EINVAL;
 	    goto bad;
 	}
@@ -1150,7 +1174,7 @@ set_perms(int perm)
 	state->rgid = getgid();
 	state->egid = getegid();
 	state->grlist = user_group_list;
-	grlist_addref(state->grlist);
+	sudo_grlist_addref(state->grlist);
 	sudo_debug_printf(SUDO_DEBUG_INFO, "%s: PERM_INITIAL: "
 	    "ruid: %d, euid: %d, rgid: %d, egid: %d", __func__,
 	    (int)state->ruid, (int)state->euid,
@@ -1165,9 +1189,16 @@ set_perms(int perm)
 	state->ruid = ROOT_UID;
 	state->euid = ROOT_UID;
 	state->rgid = ostate->rgid;
-	state->egid = ostate->egid;
+	state->egid = ROOT_GID;
+	sudo_debug_printf(SUDO_DEBUG_INFO, "%s: PERM_ROOT: gid: "
+	    "[%d, %d] -> [%d, %d]", __func__, (int)ostate->rgid,
+	    (int)ostate->egid, ROOT_GID, ROOT_GID);
+	if (GID_CHANGED && setegid(ROOT_GID)) {
+	    errstr = N_("unable to change to root gid");
+	    goto bad;
+	}
 	state->grlist = ostate->grlist;
-	grlist_addref(state->grlist);
+	sudo_grlist_addref(state->grlist);
 	break;
 
     case PERM_USER:
@@ -1182,10 +1213,10 @@ set_perms(int perm)
 	    goto bad;
 	}
 	state->grlist = user_group_list;
-	grlist_addref(state->grlist);
+	sudo_grlist_addref(state->grlist);
 	if (state->grlist != ostate->grlist) {
 	    if (sudo_setgroups(state->grlist->ngids, state->grlist->gids)) {
-		strlcpy(errbuf, "PERM_USER: setgroups", sizeof(errbuf));
+		errstr = "PERM_USER: setgroups";
 		goto bad;
 	    }
 	}
@@ -1214,10 +1245,10 @@ set_perms(int perm)
 	    goto bad;
 	}
 	state->grlist = user_group_list;
-	grlist_addref(state->grlist);
+	sudo_grlist_addref(state->grlist);
 	if (state->grlist != ostate->grlist) {
 	    if (sudo_setgroups(state->grlist->ngids, state->grlist->gids)) {
-		strlcpy(errbuf, "PERM_FULL_USER: setgroups", sizeof(errbuf));
+		errstr = "PERM_FULL_USER: setgroups";
 		goto bad;
 	    }
 	}
@@ -1240,24 +1271,28 @@ set_perms(int perm)
 	    "[%d, %d] -> [%d, %d]", __func__, (int)ostate->rgid,
 	    (int)ostate->egid, (int)state->rgid, (int)state->egid);
 	if (GID_CHANGED && setegid(state->egid)) {
-	    strlcpy(errbuf, _("unable to change to runas gid"), sizeof(errbuf));
+	    errstr = N_("unable to change to runas gid");
 	    goto bad;
 	}
 	state->grlist = runas_setgroups();
+	if (state->grlist == NULL) {
+	    errstr = N_("unable to set runas group vector");
+	    goto bad;
+	}
 	state->ruid = ostate->ruid;
 	state->euid = runas_pw ? runas_pw->pw_uid : user_uid;
 	sudo_debug_printf(SUDO_DEBUG_INFO, "%s: PERM_RUNAS: uid: "
 	    "[%d, %d] -> [%d, %d]", __func__, (int)ostate->ruid,
 	    (int)ostate->euid, (int)state->ruid, (int)state->euid);
 	if (seteuid(state->euid)) {
-	    strlcpy(errbuf, _("unable to change to runas uid"), sizeof(errbuf));
+	    errstr = N_("unable to change to runas uid");
 	    goto bad;
 	}
 	break;
 
     case PERM_SUDOERS:
 	state->grlist = ostate->grlist;
-	grlist_addref(state->grlist);
+	sudo_grlist_addref(state->grlist);
 
 	/* assume euid == ROOT_UID, ruid == user */
 	state->rgid = ostate->rgid;
@@ -1266,7 +1301,7 @@ set_perms(int perm)
 	    "[%d, %d] -> [%d, %d]", __func__, (int)ostate->rgid,
 	    (int)ostate->egid, (int)state->rgid, (int)state->egid);
 	if (GID_CHANGED && setegid(sudoers_gid)) {
-	    strlcpy(errbuf, _("unable to change to sudoers gid"), sizeof(errbuf));
+	    errstr = N_("unable to change to sudoers gid");
 	    goto bad;
 	}
 
@@ -1292,7 +1327,7 @@ set_perms(int perm)
 
     case PERM_TIMESTAMP:
 	state->grlist = ostate->grlist;
-	grlist_addref(state->grlist);
+	sudo_grlist_addref(state->grlist);
 	state->rgid = ostate->rgid;
 	state->egid = ostate->egid;
 	state->ruid = ROOT_UID;
@@ -1309,23 +1344,25 @@ set_perms(int perm)
     }
 
     perm_stack_depth++;
-    debug_return_bool(1);
+    debug_return_bool(true);
 bad:
-    warningx("%s: %s", errbuf,
-	errno == EAGAIN ? _("too many processes") : strerror(errno));
-    if (noexit)
-	debug_return_bool(0);
-    exit(1);
+    if (errno == EAGAIN)
+	sudo_warnx(U_("%s: %s"), U_(errstr), U_("too many processes"));
+    else
+	sudo_warn("%s", U_(errstr));
+    debug_return_bool(false);
 }
 
-void
+bool
 restore_perms(void)
 {
     struct perm_state *state, *ostate;
-    debug_decl(restore_perms, SUDO_DEBUG_PERMS)
+    debug_decl(restore_perms, SUDOERS_DEBUG_PERMS)
 
-    if (perm_stack_depth < 2)
-	debug_return;
+    if (perm_stack_depth < 2) {
+	sudo_warnx(U_("perm stack underflow"));
+	debug_return_bool(true);
+    }
 
     state = &perm_stack[perm_stack_depth - 1];
     ostate = &perm_stack[perm_stack_depth - 2];
@@ -1344,34 +1381,34 @@ restore_perms(void)
      * real and effective uids to ROOT_UID initially to be safe.
      */
     if (seteuid(ROOT_UID)) {
-	warningx("seteuid() [%d] -> [%d]", (int)state->euid, ROOT_UID);
+	sudo_warnx("seteuid() [%d] -> [%d]", (int)state->euid, ROOT_UID);
 	goto bad;
     }
     if (setuid(ROOT_UID)) {
-	warningx("setuid() [%d, %d] -> [%d, %d]", (int)state->ruid, ROOT_UID,
+	sudo_warnx("setuid() [%d, %d] -> [%d, %d]", (int)state->ruid, ROOT_UID,
 	    ROOT_UID, ROOT_UID);
 	goto bad;
     }
 
-    if (OID(egid) != -1 && setegid(ostate->egid)) {
-	warning("setegid(%d)", (int)ostate->egid);
+    if (OID(egid) != (gid_t)-1 && setegid(ostate->egid)) {
+	sudo_warn("setegid(%d)", (int)ostate->egid);
 	goto bad;
     }
     if (state->grlist != ostate->grlist) {
 	if (sudo_setgroups(ostate->grlist->ngids, ostate->grlist->gids)) {
-	    warning("setgroups()");
+	    sudo_warn("setgroups()");
 	    goto bad;
 	}
     }
-    if (OID(euid) != -1 && seteuid(ostate->euid)) {
-	warning("seteuid(%d)", ostate->euid);
+    if (OID(euid) != (uid_t)-1 && seteuid(ostate->euid)) {
+	sudo_warn("seteuid(%d)", ostate->euid);
 	goto bad;
     }
-    grlist_delref(state->grlist);
-    debug_return;
+    sudo_grlist_delref(state->grlist);
+    debug_return_bool(true);
 
 bad:
-    exit(1);
+    debug_return_bool(false);
 }
 
 #else /* !HAVE_SETRESUID && !HAVE_SETREUID && !HAVE_SETEUID */
@@ -1381,19 +1418,16 @@ bad:
  * NOTE: does not support the "stay_setuid" or timestampowner options.
  *       Also, sudoers_uid and sudoers_gid are not used.
  */
-int
+bool
 set_perms(int perm)
 {
     struct perm_state *state, *ostate = NULL;
     char errbuf[1024];
-    int noexit;
-    debug_decl(set_perms, SUDO_DEBUG_PERMS)
-
-    noexit = ISSET(perm, PERM_NOEXIT);
-    CLR(perm, PERM_MASK);
+    const char *errstr = errbuf;
+    debug_decl(set_perms, SUDOERS_DEBUG_PERMS)
 
     if (perm_stack_depth == PERM_STACK_MAX) {
-	strlcpy(errbuf, _("perm stack overflow"), sizeof(errbuf));
+	errstr = N_("perm stack overflow");
 	errno = EINVAL;
 	goto bad;
     }
@@ -1401,7 +1435,7 @@ set_perms(int perm)
     state = &perm_stack[perm_stack_depth];
     if (perm != PERM_INITIAL) {
 	if (perm_stack_depth == 0) {
-	    strlcpy(errbuf, _("perm stack underflow"), sizeof(errbuf));
+	    errstr = N_("perm stack underflow");
 	    errno = EINVAL;
 	    goto bad;
 	}
@@ -1414,39 +1448,45 @@ set_perms(int perm)
 	state->ruid = geteuid() == ROOT_UID ? ROOT_UID : getuid();
 	state->rgid = getgid();
 	state->grlist = user_group_list;
-	grlist_addref(state->grlist);
+	sudo_grlist_addref(state->grlist);
 	sudo_debug_printf(SUDO_DEBUG_INFO, "%s: PERM_INITIAL: "
 	    "ruid: %d, rgid: %d", __func__, (int)state->ruid, (int)state->rgid);
 	break;
 
     case PERM_ROOT:
 	state->ruid = ROOT_UID;
-	state->rgid = ostate->rgid;
+	state->rgid = ROOT_GID;
 	state->grlist = ostate->grlist;
-	grlist_addref(state->grlist);
+	sudo_grlist_addref(state->grlist);
 	sudo_debug_printf(SUDO_DEBUG_INFO, "%s: PERM_ROOT: uid: "
 	    "[%d] -> [%d]", __func__, (int)ostate->ruid, (int)state->ruid);
 	if (setuid(ROOT_UID)) {
 	    snprintf(errbuf, sizeof(errbuf), "PERM_ROOT: setuid(%d)", ROOT_UID);
 	    goto bad;
 	}
+	sudo_debug_printf(SUDO_DEBUG_INFO, "%s: PERM_ROOT: gid: "
+	    "[%d] -> [%d]", __func__, (int)ostate->rgid, (int)state->rgid);
+	if (setgid(ROOT_GID)) {
+	    errstr = N_("unable to change to root gid");
+	    goto bad;
+	}
 	break;
 
     case PERM_FULL_USER:
 	state->rgid = user_gid;
-	sudo_debug_printf(SUDO_DEBUG_INFO, "%s: PERM_ROOT: gid: "
+	sudo_debug_printf(SUDO_DEBUG_INFO, "%s: PERM_FULL_USER: gid: "
 	    "[%d] -> [%d]", __func__, (int)ostate->rgid, (int)state->rgid);
 	(void) setgid(user_gid);
 	state->grlist = user_group_list;
-	grlist_addref(state->grlist);
+	sudo_grlist_addref(state->grlist);
 	if (state->grlist != ostate->grlist) {
 	    if (sudo_setgroups(state->grlist->ngids, state->grlist->gids)) {
-		strlcpy(errbuf, "PERM_FULL_USER: setgroups", sizeof(errbuf));
+		errstr = "PERM_FULL_USER: setgroups";
 		goto bad;
 	    }
 	}
 	state->ruid = user_uid;
-	sudo_debug_printf(SUDO_DEBUG_INFO, "%s: PERM_ROOT: uid: "
+	sudo_debug_printf(SUDO_DEBUG_INFO, "%s: PERM_FULL_USER: uid: "
 	    "[%d] -> [%d]", __func__, (int)ostate->ruid, (int)state->ruid);
 	if (setuid(user_uid)) {
 	    snprintf(errbuf, sizeof(errbuf),
@@ -1463,28 +1503,30 @@ set_perms(int perm)
 	state->ruid = ostate->ruid;
 	state->rgid = ostate->rgid;
 	state->grlist = ostate->grlist;
-	grlist_addref(state->grlist);
+	sudo_grlist_addref(state->grlist);
 	break;
     }
 
     perm_stack_depth++;
-    debug_return_bool(1);
+    debug_return_bool(true);
 bad:
-    warningx("%s: %s", errbuf,
-	errno == EAGAIN ? _("too many processes") : strerror(errno));
-    if (noexit)
-	debug_return_bool(0);
-    exit(1);
+    if (errno == EAGAIN)
+	sudo_warnx(U_("%s: %s"), U_(errstr), U_("too many processes"));
+    else
+	sudo_warn("%s", U_(errstr));
+    debug_return_bool(false);
 }
 
-void
+boll
 restore_perms(void)
 {
     struct perm_state *state, *ostate;
-    debug_decl(restore_perms, SUDO_DEBUG_PERMS)
+    debug_decl(restore_perms, SUDOERS_DEBUG_PERMS)
 
-    if (perm_stack_depth < 2)
-	debug_return;
+    if (perm_stack_depth < 2) {
+	sudo_warnx(U_("perm stack underflow"));
+	debug_return_bool(true);
+    }
 
     state = &perm_stack[perm_stack_depth - 1];
     ostate = &perm_stack[perm_stack_depth - 2];
@@ -1495,25 +1537,25 @@ restore_perms(void)
     sudo_debug_printf(SUDO_DEBUG_INFO, "%s: gid: [%d] -> [%d]",
 	__func__, (int)state->rgid, (int)ostate->rgid);
 
-    if (OID(rgid) != -1 && setgid(ostate->rgid)) {
-	warning("setgid(%d)", (int)ostate->rgid);
+    if (OID(rgid) != (gid_t)-1 && setgid(ostate->rgid)) {
+	sudo_warn("setgid(%d)", (int)ostate->rgid);
 	goto bad;
     }
     if (state->grlist != ostate->grlist) {
 	if (sudo_setgroups(ostate->grlist->ngids, ostate->grlist->gids)) {
-	    warning("setgroups()");
+	    sudo_warn("setgroups()");
 	    goto bad;
 	}
     }
-    grlist_delref(state->grlist);
-    if (OID(ruid) != -1 && setuid(ostate->ruid)) {
-	warning("setuid(%d)", (int)ostate->ruid);
+    sudo_grlist_delref(state->grlist);
+    if (OID(ruid) != (uid_t)-1 && setuid(ostate->ruid)) {
+	sudo_warn("setuid(%d)", (int)ostate->ruid);
 	goto bad;
     }
-    debug_return;
+    debug_return_bool(true);
 
 bad:
-    exit(1);
+    debug_return_bool(false);
 }
 #endif /* HAVE_SETRESUID || HAVE_SETREUID || HAVE_SETEUID */
 
@@ -1523,23 +1565,21 @@ runas_setgroups(void)
 {
     struct passwd *pw;
     struct group_list *grlist;
-    debug_decl(runas_setgroups, SUDO_DEBUG_PERMS)
+    debug_decl(runas_setgroups, SUDOERS_DEBUG_PERMS)
 
     if (def_preserve_groups) {
-	grlist_addref(user_group_list);
+	sudo_grlist_addref(user_group_list);
 	debug_return_ptr(user_group_list);
     }
 
     pw = runas_pw ? runas_pw : sudo_user.pw;
-#ifdef HAVE_SETAUTHDB
-    aix_setauthdb(pw->pw_name);
-#endif
-    grlist = get_group_list(pw);
-#ifdef HAVE_SETAUTHDB
-    aix_restoreauthdb();
-#endif
-    if (sudo_setgroups(grlist->ngids, grlist->gids) < 0)
-	log_fatal(USE_ERRNO|MSG_ONLY, _("unable to set runas group vector"));
+    grlist = sudo_get_grlist(pw);
+    if (grlist != NULL) {
+	if (sudo_setgroups(grlist->ngids, grlist->gids) < 0) {
+	    sudo_grlist_delref(grlist);
+	    grlist = NULL;
+	}
+    }
     debug_return_ptr(grlist);
 }
 #endif /* HAVE_SETRESUID || HAVE_SETREUID || HAVE_SETEUID */
