@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 1993-1996, 1998-2005, 2007-2011
+ * Copyright (c) 1993-1996, 1998-2005, 2007-2016
  *	Todd C. Miller <Todd.Miller@courtesan.com>
  *
  * Permission to use, copy, modify, and distribute this software for any
@@ -19,8 +19,8 @@
  * Materiel Command, USAF, under agreement number F39502-99-1-0512.
  */
 
-#ifndef _SUDO_SUDOERS_H
-#define _SUDO_SUDOERS_H
+#ifndef SUDOERS_SUDOERS_H
+#define SUDOERS_SUDOERS_H
 
 #include <limits.h>
 #ifdef HAVE_STDBOOL_H
@@ -29,20 +29,20 @@
 # include "compat/stdbool.h"
 #endif /* HAVE_STDBOOL_H */
 
+#define DEFAULT_TEXT_DOMAIN	"sudoers"
+#include "sudo_gettext.h"	/* must be included before sudo_compat.h */
+
 #include <pathnames.h>
-#include "missing.h"
-#include "error.h"
-#include "alloc.h"
-#include "list.h"
-#include "fileops.h"
+#include "sudo_compat.h"
+#include "sudo_fatal.h"
+#include "sudo_queue.h"
 #include "defaults.h"
 #include "logging.h"
 #include "sudo_nss.h"
 #include "sudo_plugin.h"
-#include "sudo_debug.h"
-
-#define DEFAULT_TEXT_DOMAIN	"sudoers"
-#include "gettext.h"
+#include "sudo_conf.h"
+#include "sudo_util.h"
+#include "sudoers_debug.h"
 
 /*
  * Password db and supplementary group IDs with associated group names.
@@ -68,6 +68,8 @@ struct sudo_user {
     char *ttypath;
     char *host;
     char *shost;
+    char *runhost;
+    char *srunhost;
     char *prompt;
     char *cmnd;
     char *cmnd_args;
@@ -85,14 +87,17 @@ struct sudo_user {
     char *privs;
     char *limitprivs;
 #endif
-    char *cwd;
+    const char *cwd;
     char *iolog_file;
     GETGROUPS_T *gids;
+    int   execfd;
     int   ngids;
     int   closefrom;
     int   lines;
     int   cols;
     int   flags;
+    int   max_groups;
+    mode_t umask;
     uid_t uid;
     uid_t gid;
     pid_t sid;
@@ -108,10 +113,9 @@ struct sudo_user {
  * Return values for sudoers_lookup(), also used as arguments for log_auth()
  * Note: cannot use '0' as a value here.
  */
-/* XXX - VALIDATE_SUCCESS and VALIDATE_FAILURE instead? */
-#define VALIDATE_ERROR          0x001
-#define VALIDATE_OK		0x002
-#define VALIDATE_NOT_OK		0x004
+#define VALIDATE_ERROR		0x001
+#define VALIDATE_SUCCESS	0x002
+#define VALIDATE_FAILURE	0x004
 #define FLAG_CHECK_USER		0x010
 #define FLAG_NO_USER		0x020
 #define FLAG_NO_HOST		0x040
@@ -121,11 +125,13 @@ struct sudo_user {
 #define FLAG_AUTH_ERROR		0x400
 
 /*
- * find_path()/load_cmnd() return values
+ * find_path()/set_cmnd() return values
  */
-#define FOUND                   0
-#define NOT_FOUND               1
+#define FOUND			0
+#define NOT_FOUND		1
 #define NOT_FOUND_DOT		2
+#define NOT_FOUND_ERROR		3
+#define NOT_FOUND_PATH		4
 
 /*
  * Various modes sudo can be in (based on arguments) in hex
@@ -139,7 +145,7 @@ struct sudo_user {
 #define MODE_HELP		0x00000040
 #define MODE_LIST		0x00000080
 #define MODE_CHECK		0x00000100
-#define MODE_LISTDEFS		0x00000200
+#define MODE_ERROR		0x00000200
 #define MODE_MASK		0x0000ffff
 
 /* Mode flags */
@@ -163,8 +169,6 @@ struct sudo_user {
 #define PERM_SUDOERS             0x04
 #define PERM_RUNAS               0x05
 #define PERM_TIMESTAMP           0x06
-#define PERM_NOEXIT              0x10 /* flag */
-#define PERM_MASK                0xf0
 
 /*
  * Shortcuts for sudo_user contents.
@@ -173,6 +177,7 @@ struct sudo_user {
 #define user_uid		(sudo_user.uid)
 #define user_gid		(sudo_user.gid)
 #define user_sid		(sudo_user.sid)
+#define user_umask		(sudo_user.umask)
 #define user_passwd		(sudo_user.pw->pw_passwd)
 #define user_dir		(sudo_user.pw->pw_dir)
 #define user_gids		(sudo_user.gids)
@@ -189,8 +194,11 @@ struct sudo_user {
 #define user_prompt		(sudo_user.prompt)
 #define user_host		(sudo_user.host)
 #define user_shost		(sudo_user.shost)
+#define user_runhost		(sudo_user.runhost)
+#define user_srunhost		(sudo_user.srunhost)
 #define user_ccname		(sudo_user.krb5_ccname)
 #define safe_cmnd		(sudo_user.cmnd_safe)
+#define cmnd_fd			(sudo_user.execfd)
 #define login_class		(sudo_user.class_name)
 #define runas_pw		(sudo_user._runas_pw)
 #define runas_gr		(sudo_user._runas_gr)
@@ -201,20 +209,13 @@ struct sudo_user {
 #define	runas_limitprivs	(sudo_user.limitprivs)
 
 #ifdef __TANDEM
-# define ROOT_UID       65535
+# define ROOT_UID	65535
 #else
-# define ROOT_UID       0
+# define ROOT_UID	0
 #endif
+#define ROOT_GID	0
 
-/*
- * We used to use the system definition of PASS_MAX or _PASSWD_LEN,
- * but that caused problems with various alternate authentication
- * methods.  So, we just define our own and assume that it is >= the
- * system max.
- */
-#define SUDO_PASS_MAX	256
-
-struct lbuf;
+struct sudo_lbuf;
 struct passwd;
 struct stat;
 struct timeval;
@@ -222,52 +223,57 @@ struct timeval;
 /*
  * Function prototypes
  */
-#define YY_DECL int yylex(void)
+#define YY_DECL int sudoerslex(void)
 
 /* goodpath.c */
-bool sudo_goodpath(const char *, struct stat *);
+bool sudo_goodpath(const char *path, struct stat *sbp);
 
 /* findpath.c */
-int find_path(char *, char **, struct stat *, char *, int);
+int find_path(const char *infile, char **outfile, struct stat *sbp,
+    const char *path, int ignore_dot, char * const *whitelist);
 
 /* check.c */
-int check_user(int, int);
-void remove_timestamp(bool);
+int check_user(int validate, int mode);
 bool user_is_exempt(void);
 
+/* prompt.c */
+char *expand_prompt(const char *old_prompt, const char *auth_user);
+
+/* timestamp.c */
+int timestamp_remove(bool unlinkit);
+
 /* sudo_auth.c */
-int verify_user(struct passwd *pw, char *prompt, int validated);
+bool sudo_auth_needs_end_session(void);
+int verify_user(struct passwd *pw, char *prompt, int validated, struct sudo_conv_callback *callback);
 int sudo_auth_begin_session(struct passwd *pw, char **user_env[]);
 int sudo_auth_end_session(struct passwd *pw);
 int sudo_auth_init(struct passwd *pw);
 int sudo_auth_cleanup(struct passwd *pw);
 
-/* parse.c */
-int sudo_file_open(struct sudo_nss *);
-int sudo_file_close(struct sudo_nss *);
-int sudo_file_setdefs(struct sudo_nss *);
-int sudo_file_lookup(struct sudo_nss *, int, int);
-int sudo_file_parse(struct sudo_nss *);
-int sudo_file_display_cmnd(struct sudo_nss *, struct passwd *);
-int sudo_file_display_defaults(struct sudo_nss *, struct passwd *, struct lbuf *);
-int sudo_file_display_bound_defaults(struct sudo_nss *, struct passwd *, struct lbuf *);
-int sudo_file_display_privs(struct sudo_nss *, struct passwd *, struct lbuf *);
-
 /* set_perms.c */
-void rewind_perms(void);
-int set_perms(int);
-void restore_perms(void);
+bool rewind_perms(void);
+bool set_perms(int);
+bool restore_perms(void);
 int pam_prep_user(struct passwd *);
 
 /* gram.y */
-int yyparse(void);
+int sudoersparse(void);
+extern char *login_style;
+extern const char *errorfile;
+extern int errorlineno;
+extern bool parse_error;
+extern bool sudoers_warnings;
 
 /* toke.l */
 YY_DECL;
+extern FILE *sudoersin;
 extern const char *sudoers_file;
+extern char *sudoers;
 extern mode_t sudoers_mode;
 extern uid_t sudoers_uid;
 extern gid_t sudoers_gid;
+extern int sudolineno;
+extern int last_token;
 
 /* defaults.c */
 void dump_defaults(void);
@@ -276,23 +282,20 @@ void dump_auth_methods(void);
 /* getspwuid.c */
 char *sudo_getepw(const struct passwd *);
 
-/* zero_bytes.c */
-void zero_bytes(volatile void *, size_t);
-
 /* sudo_nss.c */
-void display_privs(struct sudo_nss_list *, struct passwd *);
-bool display_cmnd(struct sudo_nss_list *, struct passwd *);
+int display_privs(struct sudo_nss_list *, struct passwd *);
+int display_cmnd(struct sudo_nss_list *, struct passwd *);
 
 /* pwutil.c */
 __dso_public struct group *sudo_getgrgid(gid_t);
 __dso_public struct group *sudo_getgrnam(const char *);
 __dso_public void sudo_gr_addref(struct group *);
 __dso_public void sudo_gr_delref(struct group *);
-bool user_in_group(struct passwd *, const char *);
+bool user_in_group(const struct passwd *, const char *);
 struct group *sudo_fakegrnam(const char *);
-struct group_list *sudo_get_grlist(struct passwd *pw);
+struct group_list *sudo_get_grlist(const struct passwd *pw);
 struct passwd *sudo_fakepwnam(const char *, gid_t);
-struct passwd *sudo_fakepwnamid(const char *user, uid_t uid, gid_t gid);
+struct passwd *sudo_mkpwent(const char *user, uid_t uid, gid_t gid, const char *home, const char *shell);
 struct passwd *sudo_getpwnam(const char *);
 struct passwd *sudo_getpwuid(uid_t);
 void sudo_endgrent(void);
@@ -302,21 +305,20 @@ void sudo_grlist_addref(struct group_list *);
 void sudo_grlist_delref(struct group_list *);
 void sudo_pw_addref(struct passwd *);
 void sudo_pw_delref(struct passwd *);
-void sudo_setgrent(void);
-void sudo_setpwent(void);
+int  sudo_set_grlist(struct passwd *pw, char * const *groups, char * const *gids);
+int sudo_setgrent(void);
+int sudo_setpwent(void);
 void sudo_setspent(void);
 
 /* timestr.c */
 char *get_timestr(time_t, int);
 
-/* atobool.c */
-int atobool(const char *str);
-
 /* boottime.c */
-int get_boottime(struct timeval *);
+bool get_boottime(struct timespec *);
 
 /* iolog.c */
-void io_nextid(char *iolog_dir, char *iolog_dir_fallback, char sessid[7]);
+bool io_nextid(char *iolog_dir, char *iolog_dir_fallback, char sessid[7]);
+bool io_set_max_sessid(const char *sessid);
 
 /* iolog_path.c */
 char *expand_iolog_path(const char *prefix, const char *dir, const char *file,
@@ -324,13 +326,14 @@ char *expand_iolog_path(const char *prefix, const char *dir, const char *file,
 
 /* env.c */
 char **env_get(void);
-void env_merge(char * const envp[], bool overwrite);
-void env_init(char * const envp[]);
-void init_envtables(void);
-void insert_env_vars(char * const envp[]);
-void read_env_file(const char *, int);
-void rebuild_env(void);
-void validate_env_vars(char * const envp[]);
+bool env_merge(char * const envp[]);
+bool env_swap_old(void);
+bool env_init(char * const envp[]);
+bool init_envtables(void);
+bool insert_env_vars(char * const envp[]);
+bool read_env_file(const char *, int);
+bool rebuild_env(void);
+bool validate_env_vars(char * const envp[]);
 int sudo_setenv(const char *var, const char *val, int overwrite);
 int sudo_unsetenv(const char *var);
 char *sudo_getenv(const char *name);
@@ -339,28 +342,11 @@ int sudoers_hook_putenv(char *string, void *closure);
 int sudoers_hook_setenv(const char *name, const char *value, int overwrite, void *closure);
 int sudoers_hook_unsetenv(const char *name, void *closure);
 
-/* fmt_string.c */
-char *fmt_string(const char *, const char *);
-
 /* sudoers.c */
-void plugin_cleanup(int);
-void set_fqdn(void);
 FILE *open_sudoers(const char *, bool, bool *);
-
-/* aix.c */
-void aix_restoreauthdb(void);
-void aix_setauthdb(char *user);
-
-/* group_plugin.c */
-int group_plugin_load(char *plugin_info);
-void group_plugin_unload(void);
-int group_plugin_query(const char *user, const char *group,
-    const struct passwd *pwd);
-
-/* setgroups.c */
-int sudo_setgroups(int ngids, const GETGROUPS_T *gids);
-
-#ifndef _SUDO_MAIN
+int sudoers_policy_init(void *info, char * const envp[]);
+int sudoers_policy_main(int argc, char * const argv[], int pwflag, char *env_add[], void *closure);
+void sudoers_cleanup(void);
 extern struct sudo_user sudo_user;
 extern struct passwd *list_pw;
 extern int long_list;
@@ -368,6 +354,37 @@ extern int sudo_mode;
 extern uid_t timestamp_uid;
 extern sudo_conv_t sudo_conv;
 extern sudo_printf_t sudo_printf;
-#endif
 
-#endif /* _SUDO_SUDOERS_H */
+/* sudoers_debug.c */
+bool sudoers_debug_parse_flags(struct sudo_conf_debug_file_list *debug_files, const char *entry);
+void sudoers_debug_register(const char *plugin_path, struct sudo_conf_debug_file_list *debug_files);
+void sudoers_debug_deregister(void);
+
+/* policy.c */
+int sudoers_policy_deserialize_info(void *v, char **runas_user, char **runas_group);
+int sudoers_policy_exec_setup(char *argv[], char *envp[], mode_t cmnd_umask, char *iolog_path, void *v);
+extern const char *path_ldap_conf;
+extern const char *path_ldap_secret;
+
+/* group_plugin.c */
+int group_plugin_load(char *plugin_info);
+void group_plugin_unload(void);
+int group_plugin_query(const char *user, const char *group,
+    const struct passwd *pwd);
+extern const char *path_plugin_dir;
+
+/* editor.c */
+char *resolve_editor(const char *ed, size_t edlen, int nfiles, char **files,
+    int *argc_out, char ***argv_out, char * const *whitelist);
+
+/* gc.c */
+enum sudoers_gc_types {
+    GC_UNKNOWN,
+    GC_VECTOR,
+    GC_PTR
+};
+bool sudoers_gc_add(enum sudoers_gc_types type, void *ptr);
+bool sudoers_gc_remove(enum sudoers_gc_types type, void *ptr);
+void sudoers_gc_init(void);
+
+#endif /* SUDOERS_SUDOERS_H */
