@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 1996, 1998-2005, 2007-2012
+ * Copyright (c) 1996, 1998-2005, 2007-2015
  *	Todd C. Miller <Todd.Miller@courtesan.com>
  *
  * Permission to use, copy, modify, and distribute this software for any
@@ -21,37 +21,20 @@
  * Materiel Command, USAF, under agreement number F39502-99-1-0512.
  */
 
-#define _SUDO_MAIN
-
 #include <config.h>
 
-#include <sys/param.h>
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <sys/socket.h>
 #include <stdio.h>
-#ifdef STDC_HEADERS
-# include <stdlib.h>
-# include <stddef.h>
-#else
-# ifdef HAVE_STDLIB_H
-#  include <stdlib.h>
-# endif
-#endif /* STDC_HEADERS */
+#include <stdlib.h>
 #ifdef HAVE_STRING_H
 # include <string.h>
 #endif /* HAVE_STRING_H */
 #ifdef HAVE_STRINGS_H
 # include <strings.h>
 #endif /* HAVE_STRINGS_H */
-#ifdef HAVE_UNISTD_H
-# include <unistd.h>
-#endif /* HAVE_UNISTD_H */
-#ifdef HAVE_FNMATCH
-# include <fnmatch.h>
-#else
-# include "compat/fnmatch.h"
-#endif /* HAVE_FNMATCH */
+#include <unistd.h>
 #ifdef HAVE_NETGROUP_H
 # include <netgroup.h>
 #endif /* HAVE_NETGROUP_H */
@@ -59,19 +42,24 @@
 #include <errno.h>
 #include <netinet/in.h>
 #include <arpa/inet.h>
-#include <netdb.h>
-#ifdef HAVE_SETLOCALE
-# include <locale.h>
-#endif
 
 #include "tsgetgrpw.h"
 #include "sudoers.h"
 #include "interfaces.h"
 #include "parse.h"
 #include "sudo_conf.h"
-#include "secure_path.h"
 #include <gram.h>
 
+#ifdef HAVE_FNMATCH
+# include <fnmatch.h>
+#else
+# include "compat/fnmatch.h"
+#endif /* HAVE_FNMATCH */
+
+#ifndef YYDEBUG
+# define YYDEBUG 0
+#endif
+
 /*
  * Function Prototypes
  */
@@ -81,11 +69,9 @@ void print_defaults(void);
 void print_privilege(struct privilege *);
 void print_userspecs(void);
 void usage(void) __attribute__((__noreturn__));
-void cleanup(int);
 static void set_runaspw(const char *);
 static void set_runasgr(const char *);
-static int cb_runas_default(const char *);
-static int testsudoers_printf(int msg_type, const char *fmt, ...);
+static bool cb_runas_default(const char *);
 static int testsudoers_print(const char *msg);
 
 extern void setgrfile(const char *);
@@ -106,27 +92,19 @@ extern int (*trace_print)(const char *msg);
 /*
  * Globals
  */
-struct interface *interfaces;
 struct sudo_user sudo_user;
 struct passwd *list_pw;
 static char *runas_group, *runas_user;
-extern int errorlineno;
-extern bool parse_error;
-extern char *errorfile;
-sudo_printf_t sudo_printf = testsudoers_printf;
-sudo_conv_t sudo_conv;	/* NULL in non-plugin */
-
-/* For getopt(3) */
-extern char *optarg;
-extern int optind;
 
 #if defined(SUDO_DEVEL) && defined(__OpenBSD__)
 extern char *malloc_options;
 #endif
-#ifdef YYDEBUG
-extern int yydebug;
+#if YYDEBUG
+extern int sudoersdebug;
 #endif
 
+__dso_public int main(int argc, char *argv[]);
+
 int
 main(int argc, char *argv[])
 {
@@ -134,30 +112,29 @@ main(int argc, char *argv[])
     struct privilege *priv;
     struct userspec *us;
     char *p, *grfile, *pwfile;
-    char hbuf[MAXHOSTNAMELEN + 1];
+    const char *errstr;
     int match, host_match, runas_match, cmnd_match;
     int ch, dflag, exitcode = 0;
-    debug_decl(main, SUDO_DEBUG_MAIN)
+    debug_decl(main, SUDOERS_DEBUG_MAIN)
 
 #if defined(SUDO_DEVEL) && defined(__OpenBSD__)
     malloc_options = "AFGJPR";
 #endif
-#ifdef YYDEBUG
-    yydebug = 1;
+#if YYDEBUG
+    sudoersdebug = 1;
 #endif
 
-#if !defined(HAVE_GETPROGNAME) && !defined(HAVE___PROGNAME)
-    setprogname(argc > 0 ? argv[0] : "testsudoers");
-#endif
+    initprogname(argc > 0 ? argv[0] : "testsudoers");
 
-#ifdef HAVE_SETLOCALE 
-    setlocale(LC_ALL, "");
-#endif
+    if (!sudoers_initlocale(setlocale(LC_ALL, ""), def_sudoers_locale))
+	sudo_fatalx(U_("%s: %s"), __func__, U_("unable to allocate memory"));
     bindtextdomain("sudoers", LOCALEDIR); /* XXX - should have own domain */
     textdomain("sudoers");
 
-    /* Read sudo.conf. */
-    sudo_conf_read();
+    /* Initialize the debug subsystem. */
+    if (sudo_conf_read(NULL, SUDO_CONF_DEBUG) == -1)
+	exit(EXIT_FAILURE);
+    sudoers_debug_register(getprogname(), sudo_conf_debug_files(getprogname()));
 
     dflag = 0;
     grfile = pwfile = NULL;
@@ -170,7 +147,9 @@ main(int argc, char *argv[])
 		user_host = optarg;
 		break;
 	    case 'G':
-		sudoers_gid = (gid_t)atoi(optarg);
+		sudoers_gid = (gid_t)sudo_strtoid(optarg, NULL, NULL, &errstr);
+		if (errstr != NULL)
+		    sudo_fatalx("group ID %s: %s", optarg, errstr);
 		break;
 	    case 'g':
 		runas_group = optarg;
@@ -185,7 +164,9 @@ main(int argc, char *argv[])
 		trace_print = testsudoers_print;
 		break;
 	    case 'U':
-		sudoers_uid = (uid_t)atoi(optarg);
+		sudoers_uid = (uid_t)sudo_strtoid(optarg, NULL, NULL, &errstr);
+		if (errstr != NULL)
+		    sudo_fatalx("user ID %s: %s", optarg, errstr);
 		break;
 	    case 'u':
 		runas_user = optarg;
@@ -203,8 +184,8 @@ main(int argc, char *argv[])
 	setgrfile(grfile);
     if (pwfile)
 	setpwfile(pwfile);
-    sudo_setpwent();
-    sudo_setgrent();
+    if (sudo_setpwent() == -1 || sudo_setgrent() == -1)
+	sudo_fatalx(U_("%s: %s"), __func__, U_("unable to allocate memory"));
 
     if (argc < 2) {
 	if (!dflag)
@@ -222,21 +203,22 @@ main(int argc, char *argv[])
 	argc -= 2;
     }
     if ((sudo_user.pw = sudo_getpwnam(user_name)) == NULL)
-	errorx(1, _("unknown user: %s"), user_name);
+	sudo_fatalx(U_("unknown user: %s"), user_name);
 
     if (user_host == NULL) {
-	if (gethostname(hbuf, sizeof(hbuf)) != 0)
-	    error(1, "gethostname");
-	hbuf[sizeof(hbuf) - 1] = '\0';
-	user_host = hbuf;
+	if ((user_host = sudo_gethostname()) == NULL)
+	    sudo_fatal("gethostname");
     }
     if ((p = strchr(user_host, '.'))) {
 	*p = '\0';
-	user_shost = estrdup(user_host);
+	if ((user_shost = strdup(user_host)) == NULL)
+	    sudo_fatalx(U_("%s: %s"), __func__, U_("unable to allocate memory"));
 	*p = '.';
     } else {
 	user_shost = user_host;
     }
+    user_runhost = user_host;
+    user_srunhost = user_shost;
 
     /* Fill in user_args from argv. */
     if (argc > 0) {
@@ -246,11 +228,12 @@ main(int argc, char *argv[])
 	for (size = 0, from = argv; *from; from++)
 	    size += strlen(*from) + 1;
 
-	user_args = (char *) emalloc(size);
+	if ((user_args = malloc(size)) == NULL)
+	    sudo_fatalx(U_("%s: %s"), __func__, U_("unable to allocate memory"));
 	for (to = user_args, from = argv; *from; from++) {
 	    n = strlcpy(to, *from, size - (to - user_args));
 	    if (n >= size - (to - user_args))
-		errorx(1, _("internal error, %s overflow"), "init_vars()");
+		sudo_fatalx(U_("internal error, %s overflow"), getprogname());
 	    to += n;
 	    *to++ = ' ';
 	}
@@ -258,19 +241,22 @@ main(int argc, char *argv[])
     }
 
     /* Initialize default values. */
-    init_defaults();
+    if (!init_defaults())
+	sudo_fatalx(U_("unable to initialize sudoers default values"));
 
     /* Set runas callback. */
     sudo_defs_table[I_RUNAS_DEFAULT].callback = cb_runas_default;
 
     /* Load ip addr/mask for each interface. */
-    if (get_net_ifs(&p) > 0)
-	set_interfaces(p);
+    if (get_net_ifs(&p) > 0) {
+	if (!set_interfaces(p))
+	    sudo_fatal(U_("unable to parse network address list"));
+    }
 
     /* Allocate space for data structures in the parser. */
     init_parser("sudoers", false);
 
-    if (yyparse() != 0 || parse_error) {
+    if (sudoersparse() != 0 || parse_error) {
 	parse_error = true;
 	if (errorlineno != -1)
 	    (void) printf("Parse error in %s near line %d",
@@ -312,19 +298,19 @@ main(int argc, char *argv[])
     /* This loop must match the one in sudo_file_lookup() */
     printf("\nEntries for user %s:\n", user_name);
     match = UNSPEC;
-    tq_foreach_rev(&userspecs, us) {
+    TAILQ_FOREACH_REVERSE(us, &userspecs, userspec_list, entries) {
 	if (userlist_matches(sudo_user.pw, &us->users) != ALLOW)
 	    continue;
-	tq_foreach_rev(&us->privileges, priv) {
+	TAILQ_FOREACH_REVERSE(priv, &us->privileges, privilege_list, entries) {
 	    putchar('\n');
-	    print_privilege(priv); /* XXX */
+	    print_privilege(priv);
 	    putchar('\n');
-	    host_match = hostlist_matches(&priv->hostlist);
+	    host_match = hostlist_matches(sudo_user.pw, &priv->hostlist);
 	    if (host_match == ALLOW) {
 		puts("\thost  matched");
-		tq_foreach_rev(&priv->cmndlist, cs) {
-		    runas_match = runaslist_matches(&cs->runasuserlist,
-			&cs->runasgrouplist, NULL, NULL);
+		TAILQ_FOREACH_REVERSE(cs, &priv->cmndlist, cmndspec_list, entries) {
+		    runas_match = runaslist_matches(cs->runasuserlist,
+			cs->runasgrouplist, NULL, NULL);
 		    if (runas_match == ALLOW) {
 			puts("\trunas matched");
 			cmnd_match = cmnd_matches(cs->cmnd);
@@ -350,6 +336,8 @@ main(int argc, char *argv[])
      */
     exitcode = parse_error ? 1 : (match == ALLOW ? 0 : match + 3);
 done:
+    sudo_endpwent();
+    sudo_endgrent();
     sudo_debug_exit_int(__func__, __FILE__, __LINE__, sudo_debug_subsys, exitcode);
     exit(exitcode);
 }
@@ -357,43 +345,55 @@ done:
 static void
 set_runaspw(const char *user)
 {
-    debug_decl(main, SUDO_DEBUG_UTIL)
+    struct passwd *pw = NULL;
+    debug_decl(set_runaspw, SUDOERS_DEBUG_UTIL)
 
+    if (*user == '#') {
+	const char *errstr;
+	uid_t uid = sudo_strtoid(user + 1, NULL, NULL, &errstr);
+	if (errstr == NULL) {
+	    if ((pw = sudo_getpwuid(uid)) == NULL)
+		pw = sudo_fakepwnam(user, user_gid);
+	}
+    }
+    if (pw == NULL) {
+	if ((pw = sudo_getpwnam(user)) == NULL)
+	    sudo_fatalx(U_("unknown user: %s"), user);
+    }
     if (runas_pw != NULL)
 	sudo_pw_delref(runas_pw);
-    if (*user == '#') {
-	if ((runas_pw = sudo_getpwuid(atoi(user + 1))) == NULL)
-	    runas_pw = sudo_fakepwnam(user, runas_gr ? runas_gr->gr_gid : 0);
-    } else {
-	if ((runas_pw = sudo_getpwnam(user)) == NULL)
-	    errorx(1, _("unknown user: %s"), user);
-    }
-
+    runas_pw = pw;
     debug_return;
 }
 
 static void
 set_runasgr(const char *group)
 {
-    debug_decl(main, SUDO_DEBUG_UTIL)
+    struct group *gr = NULL;
+    debug_decl(set_runasgr, SUDOERS_DEBUG_UTIL)
 
+    if (*group == '#') {
+	const char *errstr;
+	gid_t gid = sudo_strtoid(group + 1, NULL, NULL, &errstr);
+	if (errstr == NULL) {
+	    if ((gr = sudo_getgrgid(gid)) == NULL)
+		gr = sudo_fakegrnam(group);
+	}
+    }
+    if (gr == NULL) {
+	if ((gr = sudo_getgrnam(group)) == NULL)
+	    sudo_fatalx(U_("unknown group: %s"), group);
+    }
     if (runas_gr != NULL)
 	sudo_gr_delref(runas_gr);
-    if (*group == '#') {
-	if ((runas_gr = sudo_getgrgid(atoi(group + 1))) == NULL)
-	    runas_gr = sudo_fakegrnam(group);
-    } else {
-	if ((runas_gr = sudo_getgrnam(group)) == NULL)
-	    errorx(1, _("unknown group: %s"), group);
-    }
-
+    runas_gr = gr;
     debug_return;
 }
 
 /* 
  * Callback for runas_default sudoers setting.
  */
-static int
+static bool
 cb_runas_default(const char *user)
 {
     /* Only reset runaspw if user didn't specify one. */
@@ -414,19 +414,13 @@ sudo_endspent(void)
     return;
 }
 
-void
-set_fqdn(void)
-{
-    return;
-}
-
 FILE *
 open_sudoers(const char *sudoers, bool doedit, bool *keepopen)
 {
     struct stat sb;
     FILE *fp = NULL;
     char *sudoers_base;
-    debug_decl(open_sudoers, SUDO_DEBUG_UTIL)
+    debug_decl(open_sudoers, SUDOERS_DEBUG_UTIL)
 
     sudoers_base = strrchr(sudoers, '/');
     if (sudoers_base != NULL)
@@ -437,20 +431,20 @@ open_sudoers(const char *sudoers, bool doedit, bool *keepopen)
 	    fp = fopen(sudoers, "r");
 	    break;
 	case SUDO_PATH_MISSING:
-	    warning("unable to stat %s", sudoers_base);
+	    sudo_warn("unable to stat %s", sudoers_base);
 	    break;
 	case SUDO_PATH_BAD_TYPE:
-	    warningx("%s is not a regular file", sudoers_base);
+	    sudo_warnx("%s is not a regular file", sudoers_base);
 	    break;
 	case SUDO_PATH_WRONG_OWNER:
-	    warningx("%s should be owned by uid %u",
+	    sudo_warnx("%s should be owned by uid %u",
 		sudoers_base, (unsigned int) sudoers_uid);
 	    break;
 	case SUDO_PATH_WORLD_WRITABLE:
-	    warningx("%s is world writable", sudoers_base);
+	    sudo_warnx("%s is world writable", sudoers_base);
 	    break;
 	case SUDO_PATH_GROUP_WRITABLE:
-	    warningx("%s should be owned by gid %u",
+	    sudo_warnx("%s should be owned by gid %u",
 		sudoers_base, (unsigned int) sudoers_gid);
 	    break;
 	default:
@@ -461,37 +455,29 @@ open_sudoers(const char *sudoers, bool doedit, bool *keepopen)
     debug_return_ptr(fp);
 }
 
-void
+bool
 init_envtables(void)
 {
-    return;
+    return(true);
 }
 
-int
+bool
 set_perms(int perm)
 {
-    return 1;
+    return true;
 }
 
-void
+bool
 restore_perms(void)
 {
-}
-
-void
-cleanup(int gotsignal)
-{
-    if (!gotsignal) {
-	sudo_endpwent();
-	sudo_endgrent();
-    }
+    return true;
 }
 
 void
 print_member(struct member *m)
 {
     struct sudo_command *c;
-    debug_decl(print_member, SUDO_DEBUG_UTIL)
+    debug_decl(print_member, SUDOERS_DEBUG_UTIL)
 
     if (m->negated)
 	putchar('!');
@@ -513,9 +499,9 @@ print_defaults(void)
 {
     struct defaults *d;
     struct member *m;
-    debug_decl(print_member, SUDO_DEBUG_UTIL)
+    debug_decl(print_defaults, SUDOERS_DEBUG_UTIL)
 
-    tq_foreach_fwd(&defaults, d) {
+    TAILQ_FOREACH(d, &defaults, entries) {
 	(void) fputs("Defaults", stdout);
 	switch (d->type) {
 	    case DEFAULTS_HOST:
@@ -531,8 +517,8 @@ print_defaults(void)
 		putchar('!');
 		break;
 	}
-	tq_foreach_fwd(&d->binding, m) {
-	    if (m != tq_first(&d->binding))
+	TAILQ_FOREACH(m, d->binding, entries) {
+	    if (m != TAILQ_FIRST(d->binding))
 		putchar(',');
 	    print_member(m);
 	}
@@ -552,7 +538,7 @@ print_alias(void *v1, void *v2)
     struct alias *a = (struct alias *)v1;
     struct member *m;
     struct sudo_command *c;
-    debug_decl(print_alias, SUDO_DEBUG_UTIL)
+    debug_decl(print_alias, SUDOERS_DEBUG_UTIL)
 
     switch (a->type) {
 	case HOSTALIAS:
@@ -568,8 +554,8 @@ print_alias(void *v1, void *v2)
 	    (void) printf("Runas_Alias\t%s = ", a->name);
 	    break;
     }
-    tq_foreach_fwd(&a->members, m) {
-	if (m != tq_first(&a->members))
+    TAILQ_FOREACH(m, &a->members, entries) {
+	if (m != TAILQ_FIRST(&a->members))
 	    fputs(", ", stdout);
 	if (m->type == COMMAND) {
 	    c = (struct sudo_command *) m->name;
@@ -585,70 +571,81 @@ print_alias(void *v1, void *v2)
     debug_return_int(0);
 }
 
+#define TAG_SET(tt) \
+    ((tt) != UNSPEC && (tt) != IMPLIED)
+
+#define TAG_CHANGED(t) \
+    (TAG_SET(cs->tags.t) && cs->tags.t != tags.t)
+
 void
 print_privilege(struct privilege *priv)
 {
     struct cmndspec *cs;
     struct member *m;
-    struct privilege *p;
     struct cmndtag tags;
-    debug_decl(print_privilege, SUDO_DEBUG_UTIL)
+    debug_decl(print_privilege, SUDOERS_DEBUG_UTIL)
 
-    for (p = priv; p != NULL; p = p->next) {
-	if (p != priv)
-	    fputs(" : ", stdout);
-	tq_foreach_fwd(&p->hostlist, m) {
-	    if (m != tq_first(&p->hostlist))
-		fputs(", ", stdout);
-	    print_member(m);
-	}
-	fputs(" = ", stdout);
-	tags.nopasswd = tags.noexec = UNSPEC;
-	tq_foreach_fwd(&p->cmndlist, cs) {
-	    if (cs != tq_first(&p->cmndlist))
-		fputs(", ", stdout);
-	    if (!tq_empty(&cs->runasuserlist) || !tq_empty(&cs->runasgrouplist)) {
-		fputs("(", stdout);
-		if (!tq_empty(&cs->runasuserlist)) {
-		    tq_foreach_fwd(&cs->runasuserlist, m) {
-			if (m != tq_first(&cs->runasuserlist))
-			    fputs(", ", stdout);
-			print_member(m);
-		    }  
-		} else if (tq_empty(&cs->runasgrouplist)) {
-		    fputs(def_runas_default, stdout);
-		} else {
-		    fputs(sudo_user.pw->pw_name, stdout);
-		}
-		if (!tq_empty(&cs->runasgrouplist)) {
-		    fputs(" : ", stdout);
-		    tq_foreach_fwd(&cs->runasgrouplist, m) {
-			if (m != tq_first(&cs->runasgrouplist))
-			    fputs(", ", stdout);
-			print_member(m);
-		    }
-		}
-		fputs(") ", stdout);
+    TAILQ_FOREACH(m, &priv->hostlist, entries) {
+	if (m != TAILQ_FIRST(&priv->hostlist))
+	    fputs(", ", stdout);
+	print_member(m);
+    }
+    fputs(" = ", stdout);
+    TAGS_INIT(tags);
+    TAILQ_FOREACH(cs, &priv->cmndlist, entries) {
+	if (cs != TAILQ_FIRST(&priv->cmndlist))
+	    fputs(", ", stdout);
+	if (cs->runasuserlist != NULL || cs->runasgrouplist != NULL) {
+	    fputs("(", stdout);
+	    if (cs->runasuserlist != NULL) {
+		TAILQ_FOREACH(m, cs->runasuserlist, entries) {
+		    if (m != TAILQ_FIRST(cs->runasuserlist))
+			fputs(", ", stdout);
+		    print_member(m);
+		}  
+	    } else if (cs->runasgrouplist == NULL) {
+		fputs(def_runas_default, stdout);
+	    } else {
+		fputs(sudo_user.pw->pw_name, stdout);
 	    }
+	    if (cs->runasgrouplist != NULL) {
+		fputs(" : ", stdout);
+		TAILQ_FOREACH(m, cs->runasgrouplist, entries) {
+		    if (m != TAILQ_FIRST(cs->runasgrouplist))
+			fputs(", ", stdout);
+		    print_member(m);
+		}
+	    }
+	    fputs(") ", stdout);
+	}
 #ifdef HAVE_SELINUX
-	    if (cs->role)
-		printf("ROLE=%s ", cs->role);
-	    if (cs->type)
-		printf("TYPE=%s ", cs->type);
+	if (cs->role)
+	    printf("ROLE=%s ", cs->role);
+	if (cs->type)
+	    printf("TYPE=%s ", cs->type);
 #endif /* HAVE_SELINUX */
 #ifdef HAVE_PRIV_SET
-	    if (cs->privs)
-		printf("PRIVS=%s ", cs->privs);
-	    if (cs->limitprivs)
-		printf("LIMITPRIVS=%s ", cs->limitprivs);
+	if (cs->privs)
+	    printf("PRIVS=%s ", cs->privs);
+	if (cs->limitprivs)
+	    printf("LIMITPRIVS=%s ", cs->limitprivs);
 #endif /* HAVE_PRIV_SET */
-	    if (cs->tags.nopasswd != UNSPEC && cs->tags.nopasswd != tags.nopasswd)
-		printf("%sPASSWD: ", cs->tags.nopasswd ? "NO" : "");
-	    if (cs->tags.noexec != UNSPEC && cs->tags.noexec != tags.noexec)
-		printf("%sEXEC: ", cs->tags.noexec ? "NO" : "");
-	    print_member(cs->cmnd);
-	    memcpy(&tags, &cs->tags, sizeof(tags));
-	}
+	if (TAG_CHANGED(follow))
+	    printf("%sFOLLOW: ", cs->tags.follow ? "" : "NO");
+	if (TAG_CHANGED(log_input))
+	    printf("%sLOG_INPUT: ", cs->tags.log_input ? "" : "NO");
+	if (TAG_CHANGED(log_output))
+	    printf("%sLOG_OUTPUT: ", cs->tags.log_output ? "" : "NO");
+	if (TAG_CHANGED(noexec))
+	    printf("%sEXEC: ", cs->tags.noexec ? "NO" : "");
+	if (TAG_CHANGED(nopasswd))
+	    printf("%sPASSWD: ", cs->tags.nopasswd ? "NO" : "");
+	if (TAG_CHANGED(send_mail))
+	    printf("%sMAIL: ", cs->tags.send_mail ? "" : "NO");
+	if (TAG_CHANGED(setenv))
+	    printf("%sSETENV: ", cs->tags.setenv ? "" : "NO");
+	print_member(cs->cmnd);
+	memcpy(&tags, &cs->tags, sizeof(tags));
     }
     debug_return;
 }
@@ -658,51 +655,30 @@ print_userspecs(void)
 {
     struct member *m;
     struct userspec *us;
-    debug_decl(print_userspecs, SUDO_DEBUG_UTIL)
+    struct privilege *priv;
+    debug_decl(print_userspecs, SUDOERS_DEBUG_UTIL)
 
-    tq_foreach_fwd(&userspecs, us) {
-	tq_foreach_fwd(&us->users, m) {
-	    if (m != tq_first(&us->users))
+    TAILQ_FOREACH(us, &userspecs, entries) {
+	TAILQ_FOREACH(m, &us->users, entries) {
+	    if (m != TAILQ_FIRST(&us->users))
 		fputs(", ", stdout);
 	    print_member(m);
 	}
 	putchar('\t');
-	print_privilege(us->privileges.first); /* XXX */
+	TAILQ_FOREACH(priv, &us->privileges, entries) {
+	    if (priv != TAILQ_FIRST(&us->privileges))
+		fputs(" : ", stdout);
+	    print_privilege(priv);
+	}
 	putchar('\n');
     }
     debug_return;
 }
 
-static int
-testsudoers_printf(int msg_type, const char *fmt, ...)
-{
-    va_list ap;
-    FILE *fp;
-    debug_decl(testsudoers_printf, SUDO_DEBUG_UTIL)
-            
-    switch (msg_type) {
-    case SUDO_CONV_INFO_MSG:
-	fp = stdout;
-	break;
-    case SUDO_CONV_ERROR_MSG:
-	fp = stderr;
-	break;
-    default:
-	errno = EINVAL;
-	debug_return_int(-1);
-    }
-   
-    va_start(ap, fmt);
-    vfprintf(fp, fmt, ap);
-    va_end(ap);
-   
-    debug_return_int(0);
-}
-
 void
 dump_sudoers(void)
 {
-    debug_decl(dump_sudoers, SUDO_DEBUG_UTIL)
+    debug_decl(dump_sudoers, SUDOERS_DEBUG_UTIL)
 
     print_defaults();
 
@@ -723,6 +699,6 @@ static int testsudoers_print(const char *msg)
 void
 usage(void)
 {
-    (void) fprintf(stderr, "usage: %s [-dt] [-G sudoers_gid] [-g group] [-h host] [-p grfile] [-p pwfile] [-U sudoers_uid] [-u user] <user> <command> [args]\n", getprogname());
+    (void) fprintf(stderr, "usage: %s [-dt] [-G sudoers_gid] [-g group] [-h host] [-P grfile] [-p pwfile] [-U sudoers_uid] [-u user] <user> <command> [args]\n", getprogname());
     exit(1);
 }
